#define com_maroontress_libcgi_db_XML_IMPORT
#define com_maroontress_libcgi_util_Tub_IMPORT
#define std_io_BufferedOutputTube_IMPORT
#define std_io_OutputTube_IMPORT
#define std_lang_ByteArray_IMPORT
#define std_util_Pool_IMPORT

#include <stdio.h>
#include <string.h>
#include <com/maroontress/libcgi/db/XML.h>
#include <com/maroontress/libcgi/util/Tub.h>
#include <std/io/BufferedOutputTube.h>
#include <std/io/OutputTube.h>
#include <std/lang/ByteArray.h>
#include <std/util/Pool.h>

#define ConstString_new(x) {x, sizeof(x) - 1}

struct ConstString {
    const char *s;
    size_t len;
};

static const struct ConstString *
getEscaped(int c)
{
    static const struct ConstString gt = ConstString_new("&gt;");
    static const struct ConstString lt = ConstString_new("&lt;");
    static const struct ConstString amp = ConstString_new("&amp;");
    static const struct ConstString quot = ConstString_new("&quot;");

    switch (c) {
    default:
	return NULL;
    case '>':
        return &gt;
    case '<':
        return &lt;
    case '&':
        return &amp;
    case '"':
        return &quot;
    }
}

static int
writeChar(union OutputTube out, int c)
{
    const struct ConstString *rep;
    unsigned char byte;

    if ((rep = getEscaped(c)) == NULL) {
	byte = c;
        return OutputTube_writeFully(out, &byte, 1);
    }
    return OutputTube_writeFully(out, rep->s, rep->len);
}

static int
writeCharWithTub(struct Tub *tub, int c)
{
    const struct ConstString *rep;

    if ((rep = getEscaped(c)) == NULL)
        return Tub_writeByte(tub, c);
    return Tub_write(tub, rep->s, rep->len);
}

static int
fillPool(union Pool p, const char *s)
{
    union BufferedOutputTube out;
    int status;

    if ((out = BufferedOutputTube_create(Pool_createOutputTube(p))).impl
	== NULL) {
	return -1;
    }
    while ((status = writeChar(out.super.outputTube, *s)) == 0 && *s)
	++s;
    OutputTube_close(out.super.outputTube);
    return (status < 0 || *s != 0) ? -1 : 0;
}

char *
XML_createEscaped(const char *s)
{
    union Pool p;
    struct ByteArray *a;
    char *d;

    if ((p = Pool_create()).impl == NULL)
	return NULL;
    a = (fillPool(p, s) < 0) ? NULL : Pool_createByteArray(p);
    Pool_free(p);
    if (a == NULL)
	return NULL;
    d = strdup(ByteArray_getData(a));
    ByteArray_free(a);
    return d;
}

int
XML_writeEscapedToTub(struct Tub *tub, const char *s)
{
    int status;

    while (*s && (status = writeCharWithTub(tub, *s)) == 0)
	++s;
    return (*s != 0) ? -1 : 0;
}
